<?php

/**
 * @file
 * File containing class TranscoderAbstractionFactoryElasticTranscoder.
 */

use Aws\ElasticTranscoder\ElasticTranscoderClient;
use Aws\Sns\Message;
use Aws\Sns\MessageValidator;
use Aws\Sns\Exception\InvalidSnsMessageException;
use Drupal\amazons3\S3Client;

/**
 * Class that handles Elastic Transcoder transcoding.
 */
class TranscoderAbstractionFactoryElasticTranscoder extends TranscoderAbstractionFactory implements TranscoderFactoryInterface {
  protected $options = array();
  private $postbackurl;
  private $outputdestination;

  /**
   * Instantiate a TranscoderAbstractionFactoryElasticTranscoder object.
   */
  public function __construct() {
    parent::__construct();
    $this->options['api_key'] = variable_get('amazonaws_key');
    $this->options['api_secret'] = variable_get('amazonaws_secret');

    $this->postbackurl = variable_get('video_elastic_transcoder_postback', url('postback/jobs', array('absolute' => TRUE)));
    $this->outputdestination = variable_get('video_elastic_transcoder_output_destination');
  }

  /**
   * Set the input uri for the transcoding job.
   */
  public function setInput(array $file) {
    parent::setInput($file);
    $this->options['input'] = file_create_url($this->settings['input']['uri']);
  }

  /**
   * Set the options for the transcoding job.
   */
  public function setOptions(array $options) {
    // Defaults.
    $name = md5(serialize($options));
    // Fetch the Amazon Preset ID or Create a new One.
    $preset = variable_get('amazon_elastic_transcoder_preset_' . $name, FALSE);
    if (empty($preset)) {
      $description = $options['wxh'] . ' ' . $options['video_codec'] . ' ' .
        $options['video_extension'] . ' ' . $options['video_bitrate'] . 'kbps';
      $this->options['output']['preset'] = array(
        'Name' => $name,
        'Description' => $description,
        'Container' => '',
        'Audio' => array(
          'Codec' => '',
          'CodecOptions' => array(),
          'SampleRate' => '',
          'Channels' => '',
        ),
        'Video' => array(
          'Codec' => '',
          'CodecOptions' => array(),
          'KeyframesMaxDist' => '',
          'FixedGOP' => '',
          'BitRate' => '',
          'FrameRate' => '',
          'MaxFrameRate' => '',
          'MaxWidth' => '',
          'MaxHeight' => '',
          'SizingPolicy' => '',
          'PaddingPolicy' => '',
          'DisplayAspectRatio' => '',
          'Watermarks' => array(
            '0' => array(
              'Id' => '',
              'MaxWidth' => '',
              'MaxHeight' => '',
              'SizingPolicy' => '',
              'HorizontalAlign' => '',
              'VerticalAlign' => '',
              'HorizontalOffset' => '',
              'VerticalOffset' => '',
              'Opacity' => '',
              'Target' => '',
            ),
          ),
        ),
        'Thumbnails' => array(),

      );

      // Thumbnails.
      $thumbnail_count = variable_get('video_thumbnail_count', 5);
      if ($thumbnail_count > 0) {
        $size = variable_get('video_thumbnail_size', '320x240');
        list($w, $h) = explode('x', $size);
        $this->options['output']['preset']['Thumbnails']['Format'] = 'jpg';
        $this->options['output']['preset']['Thumbnails']['MaxWidth'] = $w;
        $this->options['output']['preset']['Thumbnails']['MaxHeight'] = $h;
      }

      // Watermark.
      if (empty($options['video_watermark_enabled']) || empty($options['video_watermark_fid'])) {
        unset($this->options['output']['preset']['Watermarks']);
      }

      // Load Preset Configuration.
      foreach ($options as $key => $value) {
        if (empty($value) || $value === 'none') {
          continue;
        }

        switch ($key) {
          case 'pixel_format':
          case 'video_preset':
          case 'video_quality':
          case 'video_speed':
          case 'video_upscale':
          case 'audio_quality':
          case 'one_pass':
          case 'skip_video':
          case 'skip_audio':
          case 'auto_levels':
          case 'deblock':
          case 'denoise':
          case 'default':
          case 'clip_start':
          case 'clip_length':
            // Do Nothing.
            break;

          case 'video_extension':
            $this->options['output']['preset']['Container'] = $value;
            break;

          case 'video_codec':
            $this->options['output']['preset']['Video']['Codec'] = $value;
            break;

          case 'h264_profile':
            $this->options['output']['preset']['Video']['CodecOptions']['Profile'] = $value;
            break;

          case 'vp8_profile':
            $this->options['output']['preset']['Video']['CodecOptions']['Profile'] = $value;
            break;

          case 'video_level':
            $this->options['output']['preset']['Video']['CodecOptions']['Level'] = $value;
            break;

          case 'video_max_reference_frames':
            $this->options['output']['preset']['Video']['CodecOptions']['MaxReferenceFrames'] = intval($value);
            break;

          case 'bitrate_cap':
            $this->options['output']['preset']['Video']['CodecOptions']['MaxBitRate'] = intval($value);
            break;

          case 'buffer_size':
            $this->options['output']['preset']['Video']['CodecOptions']['BufferSize'] = intval($value);
            break;

          case 'deinterlace':
            $this->options['output']['preset']['Video']['CodecOptions']['InterlacedMode'] = $value;
            break;

          case 'video_color_space_conversion_mode':
            $this->options['output']['preset']['Video']['CodecOptions']['ColorSpaceConversionMode'] = $value;
            break;

          case 'video_chroma_subsampling':
            $this->options['output']['preset']['Video']['CodecOptions']['ChromaSubsampling'] = $value;
            break;

          case 'video_loop_count':
            $this->options['output']['preset']['Video']['CodecOptions']['LoopCount'] = ($value === 'Infinite') ? $value : intval($value);
            break;

          case 'keyframe_interval':
            $this->options['output']['preset']['Video']['KeyframesMaxDist'] = $value;

            break;

          case 'video_fixed_gop':
            $this->options['output']['preset']['Video']['FixedGOP'] = $value;
            break;

          case 'video_bitrate':
            $this->options['output']['preset']['Video']['BitRate'] = $value;
            break;

          case 'frame_rate':
            $this->options['output']['preset']['Video']['FrameRate'] = $value;
            break;

          case 'max_frame_rate':
            $this->options['output']['preset']['Video']['MaxFrameRate'] = $value;
            break;

          case 'wxh':
            list($w, $h) = explode('x', $value);
            $this->options['output']['preset']['Video']['MaxWidth'] = $w;
            $this->options['output']['preset']['Video']['MaxHeight'] = $h;
            break;

          case 'video_aspectmode':
            $this->options['output']['preset']['Video']['SizingPolicy'] = $value;
            break;

          case 'video_padding_policy':
            $this->options['output']['preset']['Video']['PaddingPolicy'] = $value;
            break;

          case 'video_aspect_ratio':
            $this->options['output']['preset']['Video']['DisplayAspectRatio'] = $value;
            break;

          case 'audio_codec':
            $this->options['output']['preset']['Audio']['Codec'] = $value;
            break;

          case 'audio_profile':
            $this->options['output']['preset']['Audio']['CodecOptions']['Profile'] = $value;
            break;

          case 'audio_flac_bit_depth':
            $this->options['output']['preset']['Audio']['CodecOptions']['BitDepth'] = intval($value);
            break;

          case 'audio_pcm_bit_depth':
            $this->options['output']['preset']['Audio']['CodecOptions']['BitDepth'] = intval($value);
            break;

          case 'audio_pcm_bit_signing':
            $this->options['output']['preset']['Audio']['CodecOptions']['Signed'] = $value;
            break;

          case 'audio_pcm_bit_order':
            $this->options['output']['preset']['Audio']['CodecOptions']['BitOrder'] = $value;
            break;

          case 'audio_sample_rate':
            $this->options['output']['preset']['Audio']['SampleRate'] = ($value === 'auto') ? $value : intval($value);
            break;

          case 'audio_bitrate':
            $this->options['output']['preset']['Audio']['BitRate'] = intval($value);
            break;

          case 'audio_channels':
            $this->options['output']['preset']['Audio']['Channels'] = ($value === 'auto') ? $value : intval($value);
            break;

          case 'audio_pcm_packing_mode':
            $this->options['output']['preset']['Audio']['AudioPackingMode'] = $value;
            break;

          case 'video_watermark_id':
            $this->options['output']['preset']['Video']['Watermarks']['0']['Id'] = $value;
            break;

          case 'video_watermark_width':
            $this->options['output']['preset']['Video']['Watermarks']['0']['MaxWidth'] = $value;
            break;

          case 'video_watermark_height':
            $this->options['output']['preset']['Video']['Watermarks']['0']['MaxHeight'] = $value;
            break;

          case 'video_watermark_sizing_policy':
            $this->options['output']['preset']['Video']['Watermarks']['0']['SizingPolicy'] = $value;
            break;

          case 'video_watermark_horizontal_alignment':
            $this->options['output']['preset']['Video']['Watermarks']['0']['HorizontalAlign'] = $value;
            break;

          case 'video_watermark_x':
            $this->options['output']['preset']['Video']['Watermarks']['0']['HorizontalOffset'] = $value;
            break;

          case 'video_watermark_vertical_alignment':
            $this->options['output']['preset']['Video']['Watermarks']['0']['VerticalAlign'] = $value;
            break;

          case 'video_watermark_y':
            $this->options['output']['preset']['Video']['Watermarks']['0']['VerticalOffset'] = $value;
            break;

          case 'video_watermark_opacity':
            $this->options['output']['preset']['Video']['Watermarks']['0']['Opacity'] = $value;
            break;

          case 'video_watermark_origin':
            $this->options['output']['preset']['Video']['Watermarks']['0']['Target'] = $value;
            break;

          case 'video_thumbnails_format':
            $this->options['output']['preset']['Thumbnails']['Format'] = $value;

          case 'video_thumbnails_interval':
            $this->options['output']['preset']['Thumbnails']['Interval'] = $value;

          case 'video_thumbnails_max_width':
            $this->options['output']['preset']['Thumbnails']['MaxWidth'] = $value;

          case 'video_thumbnails_max_height':
            $this->options['output']['preset']['Thumbnails']['MaxHeight'] = $value;

          case 'video_thumbnails_sizing_policy':
            $this->options['output']['preset']['Thumbnails']['SizingPolicy'] = $value;

          case 'video_thumbnails_padding_policy':
            $this->options['output']['preset']['Thumbnails']['PaddingPolicy'] = $value;

          default:
            break;
        }
      }
      // Unset empty Options.
      if (empty($this->options['output']['preset']['Audio']['CodecOptions'])) {
        unset($this->options['output']['preset']['Audio']['CodecOptions']);
      }
      if (empty($this->options['output']['preset']['Video']['CodecOptions'])) {
        unset($this->options['output']['preset']['Video']['CodecOptions']);
      }
      if (empty($options['video_watermark_enabled'])) {
        unset($this->options['output']['preset']['Video']['Watermarks']);
      }

      watchdog('amazon_elastic_transcoder', '!preset', array('!preset' => '<pre>' . print_r($this->options['output']['preset'], TRUE) . '</pre>'));

      $region = variable_get('amazonaws_region', NULL);
      $transcoder_client = ElasticTranscoderClient::factory(array('key' => variable_get('amazonaws_key', NULL), 'secret' => variable_get('amazonaws_secret', NULL), 'region' => $region, 'default_caching_config' => file_directory_temp()));
      $create_preset_result = $transcoder_client->createPreset($this->options['output']['preset'])->toArray();
      variable_set('amazon_elastic_transcoder_preset_' . $name, $create_preset_result);
      $this->options['output']['preset'] = $create_preset_result;
    }
    else {
      $this->options['output']['preset'] = $preset;
    }

    $prefix = 'converted';
    $path_parts = explode('/',  substr($this->settings['input']['uri'], 5));
    $target_parts = array_splice($path_parts, 1);
    $output_key = end($target_parts);
    // Append Resolution to the output_key.
    list($file, $ext) = explode('.', $output_key);
    $file .= '_' . $this->options['output']['preset']['Preset']['Video']['MaxWidth'] . 'x' . $this->options['output']['preset']['Preset']['Video']['MaxHeight'];
    $output_key = $file . '.' . $ext;
    $input_path = implode('/', $target_parts);

    $this->options['output']['job'] = array(
      'PipelineId' => variable_get('amazonaws_pipeline_id', NULL),
      'Input' => array(
        'Key' => $input_path,
      ),
      'Output' => array(
        'Key' => "$prefix/$output_key",
        'PresetId' => $this->options['output']['preset']['Preset']['Id'],
      ),
      'OutputKeyPrefix' => "videos/",
    );

    if (!empty($options['video_watermark_enabled']) && !empty($options['video_watermark_fid'])) {
      $file = file_load($options['video_watermark_fid']);
      $audioonly = !empty($options['video_watermark_onlyforaudio']);
      $isaudio = strncmp($this->settings['input']['filemime'], 'audio/', 6) === 0;

      if (!empty($file) && (!$audioonly || $isaudio)) {
        $path_parts = explode('/',  substr($file->uri, 5));
        $target_parts = array_splice($path_parts, 1);
        $target_path = implode('/', $target_parts);
        $wm = array(
          'InputKey' => $target_path,
          'PresetWatermarkId' => $this->options['output']['preset']['Preset']['Video']['Watermarks'][0]['Id'],
        );
        $this->options['output']['job']['Output']['Watermarks'][] = $wm;
      }
    }

    if (!empty($options['video_thumbnails_enabled'])) {
      list($file, $ext) = explode('.', $output_key);
      $width = ($this->options['output']['preset']['Preset']['Thumbnails']['MaxWidth'] != 'auto') ?
        $this->options['output']['preset']['Preset']['Thumbnails']['MaxWidth'] :
        $this->options['output']['preset']['Preset']['Video']['MaxWidth'];
      $height = ($this->options['output']['preset']['Preset']['Thumbnails']['MaxHeight'] != 'auto') ?
        $this->options['output']['preset']['Preset']['Thumbnails']['MaxHeight'] :
        $this->options['output']['preset']['Preset']['Video']['MaxHeight'];
      $file .= '_' . $width . 'x' . $height;
      $file = str_replace(' ', '+', $file);
      $thumbnail_pattern = "$file-{count}";
      $fid = $this->settings['input']['fid'];
      $this->options['output']['job']['Output']['ThumbnailPattern'] = "$prefix/thumbnails/$fid/$thumbnail_pattern";
    }

    return TRUE;
  }

  /**
   * Set the output uri for the transcoding job.
   */
  public function setOutput($output_directory, $output_name, $overwrite_mode = FILE_EXISTS_REPLACE) {
    parent::setOutput($output_directory, $output_name, $overwrite_mode);
    $this->options['output']['label'] = 'video-' . $this->settings['input']['fid'];
    $this->options['output']['filename'] = $this->settings['filename'];
    $this->options['output']['public'] = !variable_get('', FALSE);
    $baseurl = NULL;

    if ($this->outputdestination == 's3') {
      $bucket = variable_get('amazons3_bucket');
      // For now, silently ignore the "Use Amazon S3 module" setting when the bucket is not found
      if ($bucket !== NULL) {
        $baseurl = 's3://' . $bucket . '/';
      }
    }
    elseif ($this->outputdestination == 'rcf') {
      $username = variable_get('rackspace_cloud_username');
      $key = variable_get('rackspace_cloud_api_key');
      $container = variable_get('rackspace_cloud_container');
      $authurl = variable_get('rackspace_cloud_auth_url');
      // For now, silently ignore the "Use Rackspace Cloud Files module" setting when the cloudfiles module isn't setup
      if ($username !== NULL && $key !== NULL && $container !== NULL && $authurl !== NULL) {
        $scheme = $authurl == 'https://lon.auth.api.rackspacecloud.com' ? 'cf+uk' : 'cf';
        $baseurl = $scheme . '://' . rawurlencode($username) . ':' . rawurlencode($key) . '@' . $container . '/';
      }
    }

    if ($baseurl != NULL) {
      $this->options['output']['base_url'] = $baseurl . file_uri_target($output_directory) . '/';

      if (isset($this->options['output']['thumbnails'])) {
        $this->options['output']['thumbnails']['base_url'] = $baseurl . variable_get('video_thumbnail_path', 'videos/thumbnails') . '/' . $this->settings['input']['fid'] . '/';
      }
    }
  }

  /**
   * For new videos, this function is never called, because all thumbnails are
   * extracted and saved to the databases during the post back handler in
   * TranscoderAbstractionFactoryElasticTranscoder::processPostback().
   */
  public function extractFrames($destinationScheme, $format) {
    // Check if the job has been completed.
    // If the job has not been completed, don't bother checking for thumbnails.
    $fid = $this->settings['input']['fid'];
    $job = video_jobs::load($fid);
    if (empty($job)) {
      return array();
    }
    // No thumbnails available yet.
    if ($job->video_status != VIDEO_RENDERING_COMPLETE) {
      return array();
    }

    $path = variable_get('video_thumbnail_path', 'videos/thumbnails') . '/' . $fid;

    // Get the file system directory.
    $srcuribase = $destinationScheme . '://' . $path . '/';
    file_prepare_directory($dsturibase, FILE_CREATE_DIRECTORY);
    $srcwrapper = file_stream_wrapper_get_instance_by_scheme($destinationScheme);

    $thumbscheme = !empty($field['settings']['uri_scheme_thumbnails']) ? $field['settings']['uri_scheme_thumbnails'] : 'public';
    $thumburibase = $thumbscheme . '://' . variable_get('video_thumbnail_path', 'videos/thumbnails') . '/' . $fid . '/';

    // Check to see if the transcoder job made thumbnails.
    $thumbnails = file_scan_directory($thumburibase, '/.*\.[jpg|png]/');
    if (empty($thumbnails)) {
      return array();
    }

    $thumbs = array();
    // Total thumbs to generate.
    $no_of_thumbnails = variable_get('video_thumbnail_count', 5);
    $i = 1;
    foreach ($thumbnails as $uri => $thumbnail) {
      $ext = video_utility::getExtension($thumbnail->uri);
      $file = substr($thumbnail->name, -5);
      $filename = file_munge_filename($file . sprintf('%04d', $i) . '.' . $ext, '', TRUE);
      $dsturi = $dsturibase . $filename;

      // Download file from S3, if available.
      if (!file_exists($dsturi)) {
        $srcuri = $srcuribase . $filename;
        if (!file_exists($srcuri)) {
          watchdog('transcoder',
            'Error downloading thumbnail for video %filename: %thumbpath does not exist.',
            array('%filename' => $this->settings['input']['filename'], '%thumbpath' => $srcuri),
            WATCHDOG_ERROR);
          break;
        }

        $this->moveFile($srcuri, $dsturi);

        // Delete the source, it is no longer needed.
        drupal_unlink($srcuri);
      }

      $thumb = new stdClass();
      $thumb->status = 0;
      $thumb->filename = $filename;
      $thumb->uri = $dsturi;
      $thumb->filemime = $dstwrapper->getMimeType($dsturi);
      $thumbs[] = $thumb;
      $i++;
    }

    return !empty($thumbs) ? $thumbs : FALSE;
  }

  /**
   * Start the transcoding job.
   */
  public function execute() {
    $region = variable_get('amazonaws_region', NULL);
    $transcoder_client = ElasticTranscoderClient::factory(array('key' => variable_get('amazonaws_key', NULL), 'secret' => variable_get('amazonaws_secret', NULL), 'region' => $region, 'default_caching_config' => file_directory_temp()));
    watchdog('transcoder', 'Submitting job @job', array('@job' => print_r($this->options['output']['job'], TRUE)));
    try {
      $encoding_job = $transcoder_client->createJob($this->options['output']['job'])->toArray();

      $output = new stdClass();
      $output->filename = $this->settings['filename'];
      $output->uri = $this->settings['base_url'] . '/' . $this->settings['filename'];
      $output->filesize = 0;
      $output->timestamp = time();
      $output->jobid = $encoding_job['Job']['Id'];
      $output->duration = $encoding_job['Job']['Outputs']['0']['Duration'];

      return $output;
    }
    catch (Aws\ElasticTranscoder\Exception $e) {
      $message = $e->getMessage();
      watchdog('amazon_elastic_transcoder', 'Elastic Transcoder reports errors while converting %file:<br/>!errorlist', array(
        '%file' => $this->settings['filename'],
        '!errorlist' => $message,
      ),
      WATCHDOG_ERROR);
      return FALSE;
    }
  }

  /**
   * Returns the human readable name of this transcoder.
   */
  public function getName() {
    return 'Elastic Transcoder';
  }

  /**
   * Returns the class name of this transcoder.
   */
  public function getValue() {
    return 'TranscoderAbstractionFactoryElasticTranscoder';
  }

  /**
   * Determine if Elastic Transcoder is configured properly.
   */
  public function isAvailable(&$errormsg) {
    registry_rebuild();

    if (!module_exists('amazon_elastic_transcoder')) {
      $errormsg = t('You must <a href="@url">enable the Amazon Elastic Transcoder module</a> to use Elastic Transcoder to transcode videos.', array('@url' => url('admin/modules/', array('fragment' => 'edit-modules-media-amazon-elastic-transcoder-enable'))));
      return FALSE;
    }
    elseif (!class_exists('Aws\ElasticTranscoder\ElasticTranscoderClient')) {
      $errormsg = t('The Amazon Elastic Transcoder module has not been setup properly. Make sure that the Composer dependencies are installed');
      return FALSE;
    }

    return TRUE;
  }

  /**
   * The version of this transcoder.
   */
  public function getVersion() {
    return '0.1';
  }

  /**
   * Admin Settings.
   *
   * The admin configuration form at "/admin/config/media/video/transcoders".
   */
  public function adminSettings() {
    $t = get_t();

    $form = array();
    $amazonaws_api = variable_get('amazonaws_key', NULL);
    if (empty($amazonaws_api)) {
      $form['amazonaws_key'] = array(
        '#type' => 'textfield',
        '#title' => t('Amazon AWS API Key'),
        '#default_value' => variable_get('amazonaws_key', ''),
        '#description' => t('The key for an IAM account with access to transcoding services.'),
        '#required' => TRUE,
      );

      $form['amazonaws_secret'] = array(
        '#type' => 'textfield',
        '#title' => t('Amazon AWS API Secret'),
        '#default_value' => variable_get('amazonaws_secret', ''),
        '#description' => t('The secret for an IAM account with access to transcoding services.'),
        '#required' => TRUE,
      );
    }
    else {
      // AWS API key exists.
      $form['elastic_transcoder_info'] = array(
        '#type' => 'fieldset',
        '#title' => t('Amazon Elastic Transcoder'),
        '#collapsible' => FALSE,
        '#collapsed' => FALSE,
        '#states' => array(
          'visible' => array(
            ':input[name=video_convertor]' => array('value' => 'TranscoderAbstractionFactoryElasticTranscoder'),
          ),
        ),
      );
      $form['elastic_transcoder_info']['api_status'] = array(
        '#type' => 'item',
        '#title' => t('Amazon Elastic Transcoder API status'),
        '#markup' => $this->getCurrentStatus(),
      );
      $form['elastic_transcoder_info']['amazonaws_key'] = array(
        '#type' => 'textfield',
        '#title' => t('Amazon AWS API Key'),
        '#default_value' => variable_get('amazonaws_key', NULL),
        '#description' => t('The key for an IAM account with access to transcoding services.'),
      );
      $form['elastic_transcoder_info']['amazonaws_secret'] = array(
        '#type' => 'textfield',
        '#title' => t('Amazon AWS API Secret'),
        '#default_value' => variable_get('amazonaws_secret', NULL),
        '#description' => t('Select an Amazon API Endpoint to reach the transcoding services.'),
      );
      $regions = array(
        'us-east-2' => t('US East (Ohio)'),
        'us-east-1' => t('US East (N. Virginia)'),
        'us-west-1' => t('US West (N. California)'),
        'us-west-2' => t('US West (Oregon)'),
        'ap-south-1' => t('Asia Pacific (Mumbai)'),
        'ap-northeast-2' => t('Asia Pacific (Seoul)'),
        'ap-southeast-1' => t('Asia Pacific (Singapore)'),
        'ap-southeast-2' => t('Asia Pacific (Sydney)'),
        'ap-northeast-1' => t('Asia Pacific (Tokyo)'),
        'ca-central-1' => t('Canada (Central)'),
        'eu-central-1' => t('EU (Frankfurt)'),
        'eu-west-1' => t('EU (Ireleand)'),
        'eu-west-2' => t('EU (London)'),
        'sa-east-1' => t('South America (São Paulo)'),
      );
      $form['elastic_transcoder_info']['amazonaws_region'] = array(
        '#type' => 'select',
        '#title' => t('Amazon AWS Region'),
        '#options' => $regions,
        '#default_value' => variable_get('amazonaws_region', NULL),
        '#description' => t('The secret for an IAM account with access to transcoding services.'),
      );
      $form['elastic_transcoder_info']['amazonaws_pipeline_id'] = array(
        '#type' => 'textfield',
        '#title' => t('Amazon Elastic Transcoder Pipeline ID'),
        '#default_value' => variable_get('amazonaws_pipeline_id', NULL),
        '#description' => t('The Pipeline ID from the AWS Management Console.')
      );
      $form['elastic_transcoder_info']['video_elastic_transcoder_postback'] = array(
        '#type' => 'textfield',
        '#title' => t('Postback URL for Amazon SNS'),
        '#description' =>
          t('Important: Don\'t change this if you don\'t know what you\'re doing. The Postback URL is used by the SNS service to send transcoding status notifications to Drupal.') . '<br/>' .
          t('Default: %value', array('%value' => url('postback/jobs', array('absolute' => TRUE)))),
        '#default_value' => $this->postbackurl,
      );
      $form['elastic_transcoder_info']['amazonaws_sns_progress_topic_arn'] = array(
        '#type' => 'textfield',
        '#title' => t('SNS Notification Progress Topic Arn'),
        '#default_value' => variable_get('amazonaws_sns_progress_topic_arn', NULL),
        '#description' => t('Configure notification topics for the pipeline, and record the Topic Arn from the SNS service.'),
      );
      $form['elastic_transcoder_info']['amazonaws_sns_warning_topic_arn'] = array(
        '#type' => 'textfield',
        '#title' => t('SNS Notification Warning Topic Arn'),
        '#default_value' => variable_get('amazonaws_sns_warning_topic_arn', NULL),
        '#description' => t('Configure notification topics for the pipeline, and record the Topic Arn from the SNS service.'),
      );
      $form['elastic_transcoder_info']['amazonaws_sns_completion_topic_arn'] = array(
        '#type' => 'textfield',
        '#title' => t('SNS Notification Completion Topic Arn'),
        '#default_value' => variable_get('amazonaws_sns_completion_topic_arn', NULL),
        '#description' => t('Configure notification topics for the pipeline, and record the Topic Arn from the SNS service.'),
      );
      $form['elastic_transcoder_info']['amazonaws_sns_error_topic_arn'] = array(
        '#type' => 'textfield',
        '#title' => t('SNS Notification Error Topic Arn'),
        '#default_value' => variable_get('amazonaws_sns_error_topic_arn', NULL),
        '#description' => t('Configure notification topics for the pipeline, and record the Topic Arn from the SNS service.'),
      );
      $form['elastic_transcoder_info']['video_thumbnail_count_et'] = array(
        '#type' => 'textfield',
        '#title' => t('Number of thumbnails'),
        '#description' => t('Number of thumbnails to display from video.'),
        '#default_value' => variable_get('video_thumbnail_count', 5),
        '#size' => 5,
      );
      $form['elastic_transcoder_info']['video_thumbnail_size'] = array(
        '#type' => 'select',
        '#title' => t('Default Dimension of thumbnails'),
        '#default_value' => variable_get('video_thumbnail_size', '320x240'),
        '#options' => video_utility::getDimensions(),
      );

      // Advanced.
      $form['elastic_transcoder_info']['advanced'] = array(
        '#type' => 'fieldset',
        '#title' => t('Advanced'),
        '#collapsible' => TRUE,
        '#collapsed' => TRUE,
      );

      $tempdestinations = array(
        '' => t('Default configured filesystem') . ' (' . t('default') . ')',
      );
      if (module_exists('amazons3') && variable_get('amazons3_bucket', FALSE)) {
        $tempdestinations['s3'] = t('Amazon S3 bucket %bucket', array('%bucket' => variable_get('amazons3_bucket')));
      }
      if (module_exists('cloud_files') && variable_get('rackspace_cloud_container', FALSE)) {
        $tempdestinations['rcf'] = t('Rackspace Cloud Files container %container', array('%container' => variable_get('rackspace_cloud_container')));
      }

      if (count($tempdestinations) > 1) {
        $form['elastic_transcoder_info']['advanced']['video_elastic_transcoder_output_destination'] = array(
          '#type' => 'radios',
          '#title' => t('Location for Elastic Transcoder output'),
          '#default_value' => $this->outputdestination === NULL ? '' : $this->outputdestination,
          '#options' => $tempdestinations,
          '#description' => t('Normally, Elastic Transcoder uploads its transcoded files to its own Amazon S3 bucket from which the Video module will copy the file to the final destination. Use this setting to use a different location. If the selected location is identical to the final destination, this saves resource intensive copy operations during handling of the postback. The final destination is set per video field and defaults to the public files folder.'),
        );

        if (isset($tempdestinations['s3'])) {
          $form['elastic_transcoder_info']['advanced']['video_elastic_transcoder_output_destination']['#description'] .= '<br/>' .
            t('Enable Elastic Transcoder to upload directly to your Amazon S3 bucket');
        }
      }

      if (module_exists('amazons3')) {
        $form['elastic_transcoder_info']['advanced']['video_elastic_transcoder_private'] = array(
          '#type' => 'checkbox',
          '#title' => t('Store files on Amazon S3 privately'),
          '#default_value' => variable_get('video_elastic_transcoder_private', FALSE),
          '#description' => t('Files stored privately are only accessible by visitors when <a href="@amazons3-settings">Presigned URLs</a> are enabled. These URLs expire, allow you to control access to the video files. For this setting to work, you must set %destination-setting-name to Amazon S3.', array('@amazons3-settings' => url('admin/config/media/amazons3'), '%destination-setting-name' => t('Location for Zencoder output'))),
        );
      }
    }
    return $form;
  }

  /**
   * Validate Admin Settings.
   *
   * Validation callback for Admin Settings form.
   */
  public function adminSettingsValidate($form, &$form_state) {
    $v = $form_state['values'];

    if (variable_get('amazonaws_key', FALSE)) {
      // Workaround for the use of the same variable in FFmpeg and Zencoder.
      $form_state['values']['video_thumbnail_count'] = $form_state['values']['video_thumbnail_count_et'];
      unset($form_state['values']['video_thumbnail_count_et']);
    }
  }

  /**
   * Process the Post from the API from the transcoding job.
   */
  public function processPostback() {
    watchdog('transcoder', 'Server: @server', array('@server' => print_r($_SERVER, TRUE)));
    watchdog('transcoder', 'Post: @post', array('@post' => print_r($_POST, TRUE)));
    $message_type = isset($_SERVER['HTTP_X_AMZ_SNS_MESSAGE_TYPE']) ? $_SERVER['HTTP_X_AMZ_SNS_MESSAGE_TYPE'] : NULL;
    if ($message_type == NULL) {
      echo 'This is the Elastic Transcoder notification handler. It seems to work fine.';
      return;
    }

    // Instantiate the Message and Validator.
    $message = Message::fromRawPostData();
    $validator = new MessageValidator();

    // Validate the message and log errors if invalid.
    try {
      $validator->validate($message);
    }
    catch (\Exception $e) {
      watchdog('transcoder', 'Notification from SNS (Elastic Transcoder) Message Validation Error: @exception', array('@exception' => $e->getMessage()));
      return MENU_NOT_FOUND;
    }

    watchdog('transcoder', 'Received an @message_type message type.', array('@message_type' => $message_type));
    switch ($message_type) {
      case 'Notification':
        break;

      case 'SubscriptionConfirmation':
        $valid_arns = array(
          'progress' => variable_get('amazonaws_sns_progress_topic_arn', NULL),
          'warning' => variable_get('amazonaws_sns_warning_topic_arn', NULL),
          'completion' => variable_get('amazonaws_sns_completion_topic_arn', NULL),
          'error' => variable_get('amazonaws_sns_error_topic_arn', NULL),
        );
        if (in_array($message['TopicArn'], $valid_arns)) {
          watchdog('transcoder', '@message_type Topic ARN is valid.', array('@message_type' => $message_type));
          drupal_http_request($message['SubscribeURL']);
        }
        else {
          watchdog('transcoder', '@message_type Topic ARN is NOT valid.', array('@message_type' => $message_type));
        }
        return;

      case 'UnsubscribeConfirmation':
        return;

      default:
        // Handle unknown message type.
        watchdog('transcoder', 'Received an unknown message type.');
        return;
    }

    $notification = drupal_json_decode($message['Message']);

    if (!isset($notification['jobId'])) {
      watchdog('transcoder', 'Postback received from Elastic Transcoder is missing the jobId parameter: @message', array('@message' => print_r($notification, TRUE)));
      echo 'Invalid data';
      return;
    }

    // Check output/job state.
    $jobid = $notification['jobId'];
    $video_output = db_query('SELECT vid, original_fid, output_fid FROM {video_output} WHERE job_id = :job_id', array(':job_id' => $jobid))->fetch();
    if (empty($video_output)) {
      echo 'Not found';
      return;
    }

    $fid = intval($video_output->original_fid);
    watchdog('transcoder', 'Postback received from Elastic Transcoder for fid: @fid, Elastic Transcoder job id: @jobid.', array('@fid' => $fid, '@jobid' => $jobid));

    // Find the transcoding job.
    $video = video_jobs::load($fid);
    if (empty($video)) {
      echo 'Transcoding job not found in database';
      return;
    }

    // Elastic Transcoder API 2.x and above use $message['outputs'].
    // For now, only one output is supported.
    $output = current($notification['outputs']);
    watchdog('transcoder', '@output', array('@output' => print_r($notification, TRUE)));
    // Find all error situations.
    if ($output['status'] === 'Error' || $notification['state'] === 'ERROR') {
      watchdog('transcoder', 'Video with fid @fid and job id @jobid is marked as failed because an error status was received from AWS.', array('@fid' => $fid, '@jobid' => $jobid), WATCHDOG_WARNING);
      watchdog('transcoder', 'Error code: @error_code, Details: @details', array('@error_code' => $message['errorCode'], '@details' => $message['messageDetails']));
      video_jobs::setFailed($video);
      echo 'Cancelled';
      return;
    }


    if ($notification['state'] !== 'COMPLETED') {
      echo 'Not finished';
      return;
    }

    // Move the converted video to its final destination.
    $outputfile = file_load($video_output->output_fid);
    if (empty($outputfile)) {
      echo 'Output file not found in database';
      return;
    }

    // Sometimes the long duration of the copy() call causes Zencoder
    // to timeout and retry the notification postback later.
    // We only copy the file when it doesn't exist or has a different file size.
    // file_save() invokes filesize(), which may return a cached value.
    // Clear the stat cache to get a fresh value.
    $bucket = variable_get('amazons3_bucket', NULL);
    $target = $bucket . '/' . $notification['outputKeyPrefix'] . $output['key'];
    $uri = "s3://$target";
    // Sometimes stat gets cached for streamwrapper uris.
    drupal_flush_all_caches();
    if (!file_exists($uri)) {
      watchdog('transcoder', 'While processing Elastic Transcoder postback the output file does not exist and must have already been copied.');
      return;
    }
    $file_size = filesize($uri);
    clearstatcache();
    if (!file_exists($outputfile->uri) || filesize($outputfile->uri) != $file_size) {
      if (!$this->moveFile($uri, $outputfile->uri)) {
        watchdog('transcoder', 'While processing Elastic Transcoder postback, failed to copy @source-uri to @target-uri.', array('@source-uri' => $uri, '@target-uri' => $outputfile->uri), WATCHDOG_ERROR);
        video_jobs::setFailed($video);
        echo 'Error while moving';
        return;
      }
    }

    $outputfile->filesize = $file_size;
    file_save($outputfile);

    // Actual processing of the response.
    $video->duration = $output['duration'];
    video_jobs::setCompleted($video);

    // Clear the field cache.
    // Normally, node_save() does this.
    // That function is not invoked in all cases.
    video_utility::clearEntityCache($video->entity_type, $video->entity_id);

    // If there are no thumbnails, delay execution.
    $target_thumbnail_uri = 's3://' . $bucket . '/' . $notification['outputKeyPrefix'] . 'converted/thumbnails/' . $fid;
    if (!file_exists($target_thumbnail_uri)) {
      watchdog('transcoder', 'No thumbnails for @target_uri', array('@target_uri' => $target_thumbnail_uri));
      sleep(30);
    }

    // Scan the target directory.
    $thumbnails = file_scan_directory($target_thumbnail_uri, '/.*\.[jpg|png]/');
    if (empty($thumbnails)) {
      watchdog('transcoder', 'No thumbnails for @target_uri after waiting.', array('@target_uri' => $target_thumbnail_uri));
      return;
    }

    // Find the entity to which the file belongs.
    $entity = video_utility::loadEntity($video->entity_type, $video->entity_id);
    if (empty($entity)) {
      watchdog('transcoder', 'The entity to which the transcoded video belongs can\'t be found anymore. Entity type: @entity-type, entity id: @entity-id.', array('@entity-type' => $video->entity_type, '@entity-id' => $video->entity_id), WATCHDOG_ERROR);
      return;
    }

    // The following information was saved in video_jobs::create().
    $fieldname = $video->data['field_name'];
    $field = field_info_field($fieldname);
    $langcode = $video->data['langcode'];
    $delta = $video->data['delta'];

    // Insanity checks.
    if (empty($entity->{$fieldname}[$langcode][$delta])) {
      // The field can't be found anymore. This may be a problem.
      watchdog('transcoder', 'The field to which video @filename was uploaded doesn\'t seem to exist anymore. Entity type: @entity-type, entity id: @entity-id, field name: @fieldname, field language: @langcode, delta: @delta.',
      array(
        '@filename' => $video->filename,
        '@entity-type' => $video->entity_type,
        '@entity-id' => $video->entity_id,
        '@fieldname' => $fieldname,
        '@langcode' => $langcode,
        '@delta' => $delta,
      ), WATCHDOG_WARNING);
      echo 'No field';
      return;
    }
    if ($entity->{$fieldname}[$langcode][$delta]['fid'] != $video->fid) {
      // The field does not contain the file we uploaded.
      watchdog('transcoder', 'The field to which video @filename was uploaded doesn\'t seem to contain this video anymore. Entity type: @entity-type, entity id: @entity-id, field name: @fieldname, field language: @langcode, delta: @delta.',
      array(
        '@filename' => $video->filename,
        '@entity-type' => $video->entity_type,
        '@entity-id' => $video->entity_id,
        '@fieldname' => $fieldname,
        '@langcode' => $langcode,
        '@delta' => $delta,
      ), WATCHDOG_WARNING);
      echo 'No field in entity';
      return;
    }

    // Destination of thumbnails.
    $thumbscheme = !empty($field['settings']['uri_scheme_thumbnails']) ? $field['settings']['uri_scheme_thumbnails'] : 'public';
    $thumburibase = $thumbscheme . '://' . $bucket . '/' . variable_get('video_thumbnail_path', 'videos/thumbnails') . '/' . $video->fid . '/';
    file_prepare_directory($thumburibase, FILE_CREATE_DIRECTORY);
    $thumbwrapper = file_stream_wrapper_get_instance_by_scheme($thumbscheme);

    // Turn the thumbnails into managed files.
    // Two jobs for the same video may finish simultaneously.
    // Lock here so there are no errors when inserting the files.
    if (!lock_acquire('video_elastic_transcoder_thumbnails:' . $video->fid, count($thumbnails) * 30)) {
      if (lock_sleep('video_elastic_transcoder_thumbnails:' . $video->fid, count($thumbnails) * 30)) {
        watchdog('transcoder', 'Failed to acquire lock to download thumbnails for @video-filename.', array('@video-filename' => $video->filename), WATCHDOG_ERROR);
        return;
      }
    }

    $existingthumbs = db_query('SELECT f.uri, f.fid, f.filesize FROM {file_managed} f INNER JOIN {video_thumbnails} t ON (f.fid = t.thumbnailfid) WHERE t.videofid = :fid', array(':fid' => $video->fid))->fetchAllAssoc('uri');
    $thumbs = array();
    $tnid = 0;
    foreach ($thumbnails as $thumbnail) {
      $ext = video_utility::getExtension($thumbnail->uri);
      $thumb = new stdClass();
      $thumb->uid = $outputfile->uid; // $entity may not have a uid property, so take it from the output file.
      $thumb->status = FILE_STATUS_PERMANENT;
      $thumb->filename = $thumbnail->name . '.' . $ext;
      $thumb->uri = $thumburibase . $thumbnail->filename;
      $thumb->filemime = $thumbwrapper->getMimeType($thumbnail->uri);
      $thumb->type = 'image'; // For the media module
      $thumb->filesize = filesize($thumbnail->uri);
      $thumb->timestamp = REQUEST_TIME;

      $shouldcopy = TRUE;
      if (isset($existingthumbs[$thumb->uri])) {
        // If the thumbnail has the same size in the database compared to the notification data, don't copy
        if (file_exists($thumb->uri) && $existingthumbs[$thumb->uri]->filesize == $thumb->filesize) {
          $shouldcopy = FALSE;
        }
        $thumb->fid = intval($existingthumbs[$thumb->uri]->fid);
      }
      watchdog('transcoder', 'thumb @thumbsrc to @thumbdest. !thumb', array('@thumbsrc' => $thumbnail->uri, '@thumbdest' => $thumb->uri, '!thumb' => '<pre>' . print_r($thumb, TRUE) . '</pre>'));

      if ($shouldcopy && !$this->moveFile($thumbnail->uri, $thumb->uri)) {
        watchdog('transcoder', 'Could not copy @thumbsrc to @thumbdest.', array('@thumbsrc' => $thumbnail->uri, '@thumbdest' => $thumb->uri), WATCHDOG_ERROR);
        continue;
      }

      file_save($thumb);

      // Saving to video_thumbnails and file_usage is only necessary when this is a new thumbnail
      if (!isset($existingthumbs[$thumb->uri])) {
        db_insert('video_thumbnails')->fields(array('videofid' => $video->fid, 'thumbnailfid' => $thumb->fid))->execute();
        file_usage_add($thumb, 'file', $video->entity_type, $video->entity_id);
      }

      $thumbs[$thumb->fid] = $thumb;
    }
    // Remove Thumbnail Dir.
    unlink($target_thumbnail_uri);
    lock_release('video_elastic_transcoder_thumbnails:' . $video->fid);

    // Clear the field cache. Normally, node_save() does this, but that function is not invoked in all cases
    video_utility::clearEntityCache($video->entity_type, $video->entity_id);

    // Skip setting the thumbnail if there are no thumbnails or when the current value is already valid
    $currentthumb = isset($entity->{$fieldname}[$langcode][$delta]['thumbnail']) ? intval($entity->{$fieldname}[$langcode][$delta]['thumbnail']) : 0;
    if (empty($thumbs) || isset($thumbs[$currentthumb])) {
      echo 'OK: Thumbnail already set';
      return;
    }

    // Set a random thumbnail fid on the entity and save the entity
    $entity->{$fieldname}[$langcode][$delta]['thumbnail'] = array_rand($thumbs);

    switch ($video->entity_type) {
      case 'node':
        node_save($entity);
        break;
      case 'comment':
        comment_save($entity);
        break;
      default:
        // entity_save() is supplied by the entity module
        if (function_exists('entity_save')) {
          entity_save($video->entity_type, $entity);
        }
        break;
    }

    echo 'OK';
  }

  private function moveFile($srcuri, $dsturi) {
    $unlinksrc = FALSE;

    // Move the file if the target is also s3:// or rcf://
    if ($this->outputdestination == 's3' || $this->outputdestination = 'rcf') {
      if (file_uri_scheme($srcuri) == $this->outputdestination && file_uri_scheme($dsturi) == $this->outputdestination) {
        $s3 = S3Client::factory();

        $src = file_stream_wrapper_get_instance_by_uri($srcuri);
        $dst = file_stream_wrapper_get_instance_by_uri($dsturi);

        $bucket = variable_get('amazons3_bucket', NULL);
        $source_key = str_replace($src->dirname() . '/', '', $src->getUri());
        $target_key = str_replace($dst->dirname() . '/', '', $dst->getUri());
        watchdog('transcoder', 'About to copy file in @bucket Key: @target, Source: @source', array('@bucket' => $bucket, '@target' => $target_key, '@source' => $source_key));

        $result = FALSE;
        try {
          $result = $s3->copyObject(array(
            'ACL' => 'private',
            'Bucket' => $bucket,
            'Key' => $target_key,
            'CopySource' => $s3::encodeKey($bucket . '/' . $source_key),
            'StorageClass' => 'REDUCED_REDUNDANCY',
            'MetadataDirective' => 'REPLACE',
          ));
        }
        catch (Exception $e) {
          $message = $e->getMessage();
          watchdog('transcoder', 'Copy operation failed: @errorlist', array('@errorlist' => $message));
        }
        watchdog('transcoder', 'Rename @srcuri to @dsturi: @result', array('@srcuri' => $srcuri, '@dsturi' => $dsturi, '@result' => print_r($result, TRUE)));

        // Delete src to save space.
        if ($result) {
          unlink('s3://' . $bucket . '/' . $source_key);
          clearstatcache();
          return TRUE;
        }
        return FALSE;
      }
    }

    $unlinksrc = TRUE;
    watchdog('transcoder', 'Copy @srcuri to $dsturi', array('@srcuri' => $srcuri, '@dsturi' => $dsturi));

    // Check if the file is already at the right place
    if ($srcuri === $dsturi) {
      return TRUE;
    }

    $result = copy($srcuri, $dsturi);
    if ($result && $unlinksrc) {
      unlink($srcuri);
    }

    // Clear the stat cache after the copy.
    clearstatcache();

    return $result;
  }

  /**
   * Get enabled and supporting codecs by Zencoder.
   */
  public function getCodecs() {
    $auto = t('Default for this extension');

    return array(
      'encode' => array(
        'video' => array(
          '' => $auto,
          'gif' => 'GIF',
          'H.264' => 'H.264',
          'mpeg2' => 'MPEG-2',
          'vp8' => 'VP8',
          'vp9' => 'VP9',
        ),
        'audio' => array(
          '' => $auto,
          'AAC' => 'AAC',
          'flac' => 'FLAC',
          'mp2' => 'MP2',
          'mp3' => 'MP3',
          'pcm' => 'PCM',
          'vorbis' => 'Vorbis',
        )
      ),
      'decode' => array(),
    );
  }

  public function getAvailableFormats($type = FALSE) {
    return array(
      'flac' => 'FLAC',
      'flv' => 'FLV',
      'gif' => 'GIF',
      'fmp4' => 'FMP4',
      'mp3' => 'MP3',
      'mp4' => 'MP4',
      'mpg' => 'MPG',
      'mxf' => 'MXF',
      'oga' => 'OGA',
      'ogg' => 'OGG',
      'ts' => 'TS',
      'wav' => 'WAV',
      'webm' => 'WebM',
    );
  }

  public function getPixelFormats() {
    // Elastic Transcoder doesn't support this.
    return array();
  }

  /**
   * Reset internal variables to their initial state.
   */
  public function reset($keepinput = FALSE) {
    parent::reset($keepinput);

    if (!$keepinput) {
      unset($this->options['input']);
    }
    unset($this->options['output']);
  }

  /**
   * Check the current Zencoder API status.
   */
  public function getCurrentStatus() {
    $json = drupal_http_request('http://status.zencoder.com/api/events.json');
    $message = t('All systems go');

    if ($json->code != 200) {
      $message = t('The Zencoder Status API URL cannot be retrieved: @error (@code).', array('@code' => $json->code, '@error' => empty($json->error) ? t('unknown error') : $json->error));
    }
    else {

      $status_messages = drupal_json_decode($json->data);

      if (!empty($status_messages)) {
        $last_message = reset($status_messages);
        $last_event = $last_message['event'];
        $message = nl2br(t('@status', array('@status' =>  $last_event['description'])));
      }
    }

    return $message;
  }

  /**
   * Whether the transcoder works by sending jobs to an external system.
   *
   * True for transcoders like Elastic Transcoder, false for transcoders like FFmpeg.
   */
  public function isOffSite() {
    return TRUE;
  }

}
